from typing import Union, List

from omegaconf import ListConfig
from torch import nn as nn

from monet_pytorch.nn_utils import get_activation_module, ActivationFunctionEnum

def _scalars_to_list(params: dict):
    param_names = list(params.keys())
    list_size = len(params['channels'])
    for k in param_names:
        if k not in ['input_channels', 'channels']: # not resizable parameters
            if not isinstance(params[k], (tuple, list, ListConfig)):
                params[k] = [params[k]] * list_size
    return params


def make_sequential_from_config(input_channels: int, channels: List[int], kernels: Union[List[int], int],
                                batchnorms: Union[List[bool], bool] = False, bn_affines: Union[List[bool], bool] = False,
                                paddings: Union[List[int], int] = 0, strides: Union[List[int], int] = 1,
                                activations: Union[List[ActivationFunctionEnum], ActivationFunctionEnum] = 'relu',
                                output_paddings: Union[List[int], int] = 0, conv_transposes: Union[List[bool], bool] = False,
                                return_params: bool = False, try_inplace_activation: bool = True,
                                layernorms: Union[List[bool], bool] = False, ln_affines: Union[List[bool], bool] = False):
    params = locals()
    params = _scalars_to_list(params)
    for i in range(len(params['channels'])):
        assert not (params['batchnorms'][i] and params['layernorms'][i]), \
               f'batchnorms {params["batchnorms"]}  layernorms {params["layernorms"]} ' \
               f'cannot be both true at the same index'
    for k in list(params):
        if k != 'input_channels':
            assert len(params[k]) == len(params['channels']), f'len({k}) = {len(params[k])} ' \
                                                              f'!= len(channels) = {len(params["channels"])}\n' \
                                                              f'{k} = {params[k]} - channels = {params["channels"]}'
    layers = []
    input_channels = input_channels
    layer_infos = zip(params['channels'], params['batchnorms'],
                      params['bn_affines'], params['kernels'],
                      params['strides'], params['paddings'],
                      params['activations'], params['conv_transposes'],
                      params['output_paddings'], params['layernorms'],
                      params['ln_affines'])
    it_counter = 0
    for i in range(len(params['channels'])):
        channel = params['channels'][i]
        if params['conv_transposes'][i]:
            layers.append(nn.ConvTranspose2d(input_channels, channel, params['kernels'][i],
                                             params['strides'][i], params['paddings'][i], params['output_paddings'][i]))
        else:
            layers.append(nn.Conv2d(input_channels, channel, params['kernels'][i],
                                    params['strides'][i], params['paddings'][i]))

        if params['batchnorms'][i]:
            layers.append(nn.BatchNorm2d(channel, affine=params['bn_affines'][i]))
        if params['layernorms'][i]:
            layers.append(nn.GroupNorm(1, channel, affine=params['ln_affines'][i]))
        if params['activations'][i] is not None:
            layers.append(get_activation_module(params['activations'][i], try_inplace=try_inplace_activation))
        input_channels = channel
        if params['activations'][i] == 'glu':
            input_channels = input_channels // 2
        it_counter += 1
    assert it_counter == len(channels), f"Number of iteration {it_counter} is different from number of layers {len(channels)}, " \
                                       f"try to check if any config parameter is shorted than channels length"
    if not return_params:
        return nn.Sequential(*layers)
    else:
        return nn.Sequential(*layers), params


from dataclasses import dataclass, field

import torch
import torch.distributions as dists
from torch import nn
from torch.nn import functional as F
from math import prod

from monet_pytorch.attention_net import AttentionNet
from monet_pytorch.template.encoder_decoder import EncoderNet, BroadcastDecoderNet, EncoderConfig, DecoderConfig

@dataclass(eq=False)  # for compatibility reasons
class Monet(nn.Module):
    width: int
    height: int
    latent_size: int
    num_slots: int
    num_blocks_unet: int
    beta_kl: float
    gamma: float

    encoder_config: EncoderConfig
    decoder_config: DecoderConfig

    input_channels: int = 3
    bg_sigma: float = 0.09
    fg_sigma: float = 0.11

    prior_mean: float = 0.0
    prior_std: float = 1.0
    
    channels_unet: int = 32
    mlp_size_unet: int = 128
    # TODO this can also be monet-iodine: maybe this default can lead to bugs? Mikedev answer: yes is config is set to that value
    name: str = 'monet'

    encoder: EncoderNet = field(init=False)
    decoder: BroadcastDecoderNet = field(init=False)

    def __post_init__(self):
        super().__init__()
        self.attention = AttentionNet(self.input_channels, self.num_blocks_unet, self.channels_unet, self.mlp_size_unet)
        self.encoder_config.update(width=self.width, height=self.height)
        self.encoder = EncoderNet(**self.encoder_config)
        self.decoder = BroadcastDecoderNet(**self.decoder_config)
        self.beta_orig = self.beta_kl
        self.prior_dist = dists.Normal(self.prior_mean, self.prior_std)
        self.register_buffer('lower_bound_mask', torch.FloatTensor([1e-5]))  # for backward compatibility

    def forward(self, x):
        scope_shape = list(x.shape)
        scope_shape[1] = 1
        log_scope = torch.zeros(scope_shape, device=x.device)
        log_masks = []
        for i in range(self.num_slots - 1):
            log_mask, log_scope = self.attention(x, log_scope)
            log_masks.append(log_mask)
        log_masks.append(log_scope)
        log_masks = torch.cat(log_masks, dim=1)

        slots_shape = list(x.shape)
        slots_shape.insert(1, self.num_slots)
        slots = torch.zeros(slots_shape, device=x.device)

        kl_zs, masks_pred, neg_log_p_xs, zs = self.forward_vae_slots(log_masks, slots, x)
        log_masks_pred = torch.cat(masks_pred, dim=1).log_softmax(dim=1)
        masks = log_masks.exp()
        neg_log_p_xs = torch.cat(neg_log_p_xs, dim=1)
        masks_pred = log_masks_pred.exp()

        kl_masks, loss, neg_log_p_xs = self._calc_loss(kl_zs, masks, masks_pred, neg_log_p_xs)
        return {'loss': loss,
             'neg_log_p_x': neg_log_p_xs,
             'kl_mask': kl_masks,
             'kl_latent': kl_zs,
             'z': zs,
             'mask': masks,
             'slot': slots,
             'mask_pred': masks_pred,
             'log_mask_pred': log_masks_pred.unsqueeze(2),
             'log_mask': log_masks.unsqueeze(2)}

    def forward_vae_slots(self, log_masks, slots, x):
        masks_pred = []
        neg_log_p_xs = []
        kl_zs = 0.0
        num_slots = log_masks.shape[1]
        zs = torch.zeros(x.size(0), num_slots, self.latent_size)
        for i in range(num_slots):
            log_mask = log_masks[:, i].unsqueeze(1)
            z, kl_z = self._encode(x, log_mask)
            sigma = self.bg_sigma if i == 0 else self.fg_sigma
            neg_log_p_x_masked, x_recon, mask_pred = self._decode(x, z, log_mask, sigma)
            neg_log_p_xs.append(neg_log_p_x_masked.unsqueeze(1))
            kl_zs += kl_z.mean(dim=0)
            masks_pred.append(mask_pred)
            slots[:, i] = x_recon
            zs[:, i, :] = z
        return kl_zs, masks_pred, neg_log_p_xs, zs


    def _calc_loss(self, kl_zs, masks, masks_pred, neg_log_p_xs):
        loss = 0.0
        neg_log_p_xs = - neg_log_p_xs.logsumexp(dim=1).mean(dim=0).sum()
        loss += neg_log_p_xs + self.beta_kl * kl_zs
        kl_masks = self._calc_kl_mask(masks, masks_pred)
        loss += self.gamma * kl_masks
        return kl_masks, loss, neg_log_p_xs

    def _calc_kl_mask(self, masks, masks_pred):
        bs = len(masks)
        flat_masks = masks.permute(0, 2, 3, 1)
        nrows = prod(flat_masks.shape[:-1])
        flat_masks = flat_masks.reshape(nrows, -1)
        flat_masks_pred = masks_pred.permute(0, 2, 3, 1)
        flat_masks_pred = flat_masks_pred.reshape(nrows, -1)
        flat_masks = flat_masks.clamp_min(1e-5)
        flat_masks_pred = flat_masks_pred.clamp_min(1e-5)
        d_masks = dists.Categorical(probs=flat_masks)
        d_masks_pred = dists.Categorical(probs=flat_masks_pred)
        kl_masks = dists.kl_divergence(d_masks, d_masks_pred)
        kl_masks = kl_masks.sum() / bs
        return kl_masks

    def _encode(self, x, log_mask):
        encoder_input = torch.cat((x, log_mask), 1)
        q_params = self.encoder(encoder_input)
        means = q_params[:, :self.latent_size]
        sigmas = F.softplus(q_params[:, self.latent_size:])
        latent_normal = dists.Normal(means, sigmas)
        kl_z = dists.kl_divergence(latent_normal, self.prior_dist)
        kl_z = kl_z.sum(dim=1)
        z = means + sigmas * torch.randn_like(sigmas)
        return z, kl_z

    def _decode(self, x, z, log_mask, sigma):
        decoder_output = self.decoder(z)
        x_recon = decoder_output[:, :3].sigmoid()
        mask_pred = decoder_output[:, 3].unsqueeze(1)
        dist = dists.Normal(x_recon, sigma)
        neg_log_p_x_masked = log_mask + dist.log_prob(x)
        return neg_log_p_x_masked, x_recon, mask_pred



from dataclasses import dataclass
from typing import List, Literal, Union, Dict

import torch
from torch import nn as nn
from monet_pytorch.nn_utils import get_activation_module
from monet_pytorch.template.sequential_cnn import make_sequential_from_config


class EncoderConfig(Dict):
    activation: Literal['relu', 'leakyrelu', 'elu']
    channels: List[int]
    batchnorms: List[bool]
    bn_affines: List[bool]
    kernels: List[int]
    strides: List[int]
    paddings: List[int]
    mlp_hidden_size: int
    mlp_output_size: int


class DecoderConfig(Dict):
    activation: Literal['relu', 'leakyrelu', 'elu']
    channels: List[int]
    batchnorms: List[bool]
    bn_affines: List[bool]
    kernels: List[int]
    strides: List[int]
    paddings: List[int]
    w_broadcast: int
    h_broadcast: int

@dataclass(eq=False)
class EncoderNet(nn.Module):
    width: int
    height: int

    input_channels: int
    activations: Literal['relu', 'leakyrelu', 'elu']
    channels: List[int]
    batchnorms: List[bool]
    bn_affines: List[bool]
    kernels: List[int]
    strides: List[int]
    paddings: List[int]
    mlp_hidden_size: int
    mlp_output_size: int

    def __post_init__(self):
        super().__init__()
        self.convs, params = make_sequential_from_config(self.input_channels, self.channels, self.kernels,
                                                 self.batchnorms, self.bn_affines, self.paddings,
                                                 self.strides, self.activations, return_params=True)
        width = self.width
        height = self.height
        for kernel, stride, padding in zip(params['kernels'], params['strides'], params['paddings']):
            width = (width + 2 * padding - kernel) // stride + 1
            height = (height + 2 * padding - kernel) // stride + 1

        self.mlp = nn.Sequential(
            nn.Linear(self.channels[-1] * width * height, self.mlp_hidden_size),
            get_activation_module(self.activations, try_inplace=True),
            nn.Linear(self.mlp_hidden_size, self.mlp_output_size)
        )

    def forward(self, x):
        x = self.convs(x).flatten(1)
        x = self.mlp(x)
        return x


@dataclass(eq=False)
class BroadcastDecoderNet(nn.Module):
    w_broadcast: int
    h_broadcast: int
    input_channels: int
    activations: List[Union[Literal['relu', 'leakyrelu', 'elu'], None]]
    channels: List[int]
    paddings: Union[List[int], int]
    kernels: List[int]
    batchnorms: List[bool] = False
    bn_affines: List[bool] = False
    strides: List[int] = 1

    def __post_init__(self):
        super().__init__()
        self.parse_w_h()

        self.convs = []
        self.convs = make_sequential_from_config(self.input_channels, self.channels, self.kernels,
                                                 self.batchnorms, self.bn_affines, self.paddings,
                                                 self.strides, self.activations)

        ys = torch.linspace(-1, 1, self.h_broadcast)
        xs = torch.linspace(-1, 1, self.w_broadcast)
        ys, xs = torch.meshgrid(ys, xs)
        coord_map = torch.stack((ys, xs)).unsqueeze(0)
        self.register_buffer('coord_map_const', coord_map)

    def parse_w_h(self):
        if not isinstance(self.w_broadcast, int):
            self.w_broadcast = eval(self.w_broadcast)
            assert isinstance(self.w_broadcast, int)
        if not isinstance(self.h_broadcast, int):
            self.h_broadcast = eval(self.h_broadcast)
            assert isinstance(self.h_broadcast, int)

    def forward(self, z):
        batch_size = z.shape[0]
        z_tiled = z.unsqueeze(-1).unsqueeze(-1).expand(batch_size, z.shape[1], self.h_broadcast, self.w_broadcast)
        coord_map = self.coord_map_const.expand(batch_size, 2, self.h_broadcast, self.w_broadcast)
        inp = torch.cat((z_tiled, coord_map), 1)
        result = self.convs(inp)
        return result



from math import prod
from typing import Literal, Optional

import hydra.utils
from omegaconf import OmegaConf, omegaconf

from monet_pytorch import Monet
from monet_pytorch.paths import CONFIG_MODEL, CONFIG_SPECIAL_CASES, CONFIG_DATASET

OmegaConf.register_new_resolver('prod', lambda *numbers: int(prod(float(x) for x in numbers)))
OmegaConf.register_new_resolver('add', lambda *numbers: int(sum(float(x) for x in numbers)))

_MODEL_CONFIG_VALUES = ['monet', 'monet-iodine', 'monet-lightweight']
_DATASET_CONFIG_VALUES = ['clevr_6', 'multidsprites_colored_on_grayscale',
                          'tetrominoes', 'multidsprites_colored_on_colored']


def init_monet(model: Literal['monet', 'monet-iodine', 'monet-lightweight'] = 'monet',
               dataset: Optional[Literal['clevr_6', 'multidsprites_colored_on_grayscale',
                                         'tetrominoes', 'multidsprites_colored_on_colored']] = None,
               scene_max_objects: int = 5, dataset_width: int = 64, dataset_height: int = 64) -> Monet:
    assert model in _MODEL_CONFIG_VALUES

    monet_config = OmegaConf.load(CONFIG_MODEL / f'{model}.yaml')
    if dataset is None:
        monet_config.merge_with(OmegaConf.create(f"""
        dataset:
          width: {dataset_width}
          height: {dataset_height}
          max_num_objects: {scene_max_objects}
          """))
    else:
        assert dataset in _DATASET_CONFIG_VALUES
        monet_config.merge_with(OmegaConf.load(CONFIG_DATASET / f'{dataset}.yaml'))
    if dataset is not None:
        assert dataset in _DATASET_CONFIG_VALUES
        special_case_path = CONFIG_SPECIAL_CASES / f'{model}-{dataset}.yaml'
        if special_case_path.exists():
            monet_config.merge_with(OmegaConf.load(special_case_path))

    return hydra.utils.instantiate(monet_config.model)


def init_monet_custom(monet_config: omegaconf.DictConfig) -> Monet:
    target = monet_config.model['_target_']
    assert target.startswith('monet_pytorch') and target.endswith('Monet')
    return hydra.utils.instantiate(monet_config.model)



from collections import OrderedDict
from typing import List, Union, Literal, Optional

import torch
from torch import nn

ActivationFunctionEnum = Union[Literal['relu', 'leakyrelu', 'elu', 'glu'], None]


@torch.no_grad()
def init_trunc_normal(model, mean=0, std: float = 1):
    for name, tensor in model.named_parameters():
        if 'bias' in name:
            tensor.zero_()
        elif 'weight' in name:
            size = tensor.shape
            tmp = tensor.new_empty(size + (4,)).normal_()
            valid = (tmp < 2) & (tmp > -2)
            ind = valid.max(-1, keepdim=True)[1]
            tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
            tensor.data.mul_(std).add_(mean)


@torch.no_grad()
def init_xavier_(model):
    # print("Xavier init")
    for name, tensor in model.named_parameters():
        if name.endswith('.bias'):
            tensor.zero_()
        elif len(tensor.shape) == 1:
            pass  # silent
            # print(f"    skipped tensor '{name}' because shape="
            #       f"{tuple(tensor.shape)} and it does not end with '.bias'")
        else:
            torch.nn.init.xavier_uniform_(tensor)


def norm_gradient(model, p_norm):
    return sum(param.grad.norm(p_norm) for param in model.parameters())


def get_activation_module(act_f_name, try_inplace=True):
    if act_f_name == 'leakyrelu':
        ActF = torch.nn.LeakyReLU()
    elif act_f_name == 'elu':
        ActF = torch.nn.ELU()
    elif act_f_name == 'relu':
        ActF = torch.nn.ReLU(inplace=try_inplace)
    elif act_f_name == 'glu':
        ActF = torch.nn.GLU(dim=1)  # channel dimension in images
    elif act_f_name == 'sigmoid':
        ActF = torch.nn.Sigmoid()
    elif act_f_name == 'tanh':
        ActF = torch.nn.Tanh()
    else:
        raise ValueError(f"act_f_name = {act_f_name} not found")
    return ActF


def calc_output_shape_conv(width: int, height: int, kernels: List[int], paddings: List[int], strides: List[int]):
    for kernel, stride, padding in zip(kernels, strides, paddings):
        width = (width + 2 * padding - kernel) // stride + 1
        height = (height + 2 * padding - kernel) // stride + 1
    return width, height


def print_num_params(model: nn.Module, max_depth: Optional[int] = 4):
    """Prints overview of the model with number of parameters.

    Optionally, it groups together parameters below a certain depth in the
    module tree.

    Args:
        model (torch.nn.Module)
        max_depth (int, optional)
    """

    sep = '.'  # string separator in parameter name
    print("\n--- Trainable parameters:")
    num_params_tot = 0
    num_params_dict = OrderedDict()

    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue

        num_params = param.numel()

        if max_depth is not None:
            split = name.split(sep)
            prefix = sep.join(split[:max_depth])
        else:
            prefix = name
        if prefix not in num_params_dict:
            num_params_dict[prefix] = 0
        num_params_dict[prefix] += num_params
        num_params_tot += num_params
    for n, n_par in num_params_dict.items():
        print("{:8d}  {}".format(n_par, n))
    print("  - Total trainable parameters:", num_params_tot)
    print("---------\n")


def get_debug_information_model(model):
    hooks = []

    def printnorm(self, input, output):
        # input is a tuple of packed inputs
        # output is a Tensor. output.data is the Tensor we are interested
        print(self.__class__.__name__ + ' forward')
        print('')
        print('input: ', type(input))
        print('input[0]: ', type(input[0]))
        print('output: ', type(output))
        print('')
        print('input size:', input[0].size())
        if isinstance(output, dict):
            for name, value in output.items():
                print('output key: {} | size: {}'.format(name, value.data.size()))
                print('output key: {} |  norm:{}'.format(name, value.data.norm()))
        else:
            print('output size: {}'.format(output.data.size()))
            print('output norm:{}'.format(output.data.norm()))

        print('')

    def simplified_printnorm(self, input, output):
        # input is a tuple of packed inputs
        # output is a Tensor. output.data is the Tensor we are interested

        if isinstance(output, dict):
            print('{:>25} | {:>25} | {:>25}'.format(self.__class__.__name__, str(list(input[0].size())),
                                                    "Dictionary Output"))
            for name, value in output.items():
                print('{:>25} | {:>25} | {:>25} | {}'.format(" ", " ", str(list(value.data.size())), name))
                print('{:>25} | {:>25} | {:>25.2e} | {}'.format(" ", " ", value.data.norm().item(), name))

        else:
            print('{:>25} | {:>25} | {:>25}'.format(self.__class__.__name__, str(list(input[0].size())),
                                                    str(list(output.data.size()))))
            print('{:>25} | {:>25.2e} | {:>25.2e}'.format(" ", input[0].norm().item(), output.data.norm().item()))

        print('')

    for module in model.modules():
        hooks.append(module.register_forward_hook(simplified_printnorm))

    return hooks


def remove_hooks(hooks):
    for hook in hooks:
        hook.remove()



import torch
from torch import nn as nn
from torch.nn import LogSoftmax

from monet_pytorch.unet import UNet


class AttentionNet(nn.Module):
    def __init__(self, input_channels: int, num_blocks: int, filters: int = 32, mlp_size: int = 128):
        super().__init__()
        self.unet = UNet(input_channels, num_blocks, filters, mlp_size)
        self.log_softmax = LogSoftmax(dim=1)

    def forward(self, x, log_scope):
        inp = torch.cat((x, log_scope), 1)
        logits = self.unet(inp)
        #       logits = self.last_conv(logits)
        log_alpha = self.log_softmax(logits)
        log_mask = log_scope + log_alpha[:, 0:1]  # scope * alpha ~= log_scope + log_alpha
        log_scope += log_alpha[:, 1:2]  # scope * (1-alpha) ~= log_scope * log(1-alpha)
        return log_mask, log_scope



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['monet_pytorch', 'monet_pytorch.template']

package_data = \
{'': ['*'],
 'monet_pytorch': ['config/dataset/*',
                   'config/model/*',
                   'config/special_cases/*']}

install_requires = \
['hydra-core>=1.2.0,<2.0.0', 'path>=15.0,<=16.4.0', 'torch>=1.8,<=1.12.0']

setup_kwargs = {
    'name': 'monet-pytorch',
    'version': '1.0.0',
    'description': 'Pytorch implementation of Multi-Object Network(MONet)',
    'long_description': None,
    'author': 'mikedev',
    'author_email': 'mik3dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import torch
from torch import nn as nn
from torch.nn import Flatten, functional as F


class INConvBlock(nn.Module):
    def __init__(self, nin, nout, stride=1, instance_norm=True, act=nn.ReLU(inplace=True)):
        super(INConvBlock, self).__init__()
        self.conv = nn.Conv2d(nin, nout, 3, stride, 1, bias=not instance_norm)
        if instance_norm:
            self.instance_norm = nn.InstanceNorm2d(nout, affine=True)
        else:
            self.instance_norm = None
        self.act = act

    def forward(self, x):
        x = self.conv(x)
        if self.instance_norm is not None:
            x = self.instance_norm(x)
        return self.act(x)


class UNet(nn.Module):

    def __init__(self, input_channels: int, num_blocks, filter_start=32, mlp_size: int = 128):
        super(UNet, self).__init__()
        c = filter_start
        self.mlp_size = mlp_size
        if num_blocks == 4:
            self.down = nn.ModuleList([
                INConvBlock(input_channels + 1, c),
                INConvBlock(c, 2 * c),
                INConvBlock(2 * c, 2 * c),
                INConvBlock(2 * c, 2 * c),  # no downsampling
            ])
            self.up = nn.ModuleList([
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, c),
                INConvBlock(2 * c, c)
            ])
        elif num_blocks == 5:
            self.down = nn.ModuleList([
                INConvBlock(4, c),
                INConvBlock(c, c),
                INConvBlock(c, 2 * c),
                INConvBlock(2 * c, 2 * c),
                INConvBlock(2 * c, 2 * c),  # no downsampling
            ])
            self.up = nn.ModuleList([
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, c),
                INConvBlock(2 * c, c),
                INConvBlock(2 * c, c)
            ])
        elif num_blocks == 6:
            self.down = nn.ModuleList([
                INConvBlock(4, c),
                INConvBlock(c, c),
                INConvBlock(c, c),
                INConvBlock(c, 2 * c),
                INConvBlock(2 * c, 2 * c),
                INConvBlock(2 * c, 2 * c),  # no downsampling
            ])
            self.up = nn.ModuleList([
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, 2 * c),
                INConvBlock(4 * c, c),
                INConvBlock(2 * c, c),
                INConvBlock(2 * c, c),
                INConvBlock(2 * c, c)
            ])
        self.mlp = nn.Sequential(
            Flatten(),
            nn.Linear(4 * 4 * 2 * c, mlp_size), nn.ReLU(inplace=True),
            nn.Linear(mlp_size, mlp_size), nn.ReLU(inplace=True),
            nn.Linear(mlp_size, 4 * 4 * 2 * c), nn.ReLU(inplace=True)
        )
        self.final_conv = nn.Conv2d(c, 2, 1)

    def forward(self, x):
        batch_size = x.size(0)
        x_down = [x]
        skip = []
        for i, block in enumerate(self.down):
            act = block(x_down[-1])
            skip.append(act)
            if i < len(self.down) - 1:
                act = F.interpolate(act, scale_factor=0.5, mode='nearest', recompute_scale_factor=True)
            x_down.append(act)
        x_up = self.mlp(x_down[-1]).view(batch_size, -1, 4, 4)
        for i, block in enumerate(self.up):
            features = torch.cat([x_up, skip[-1 - i]], dim=1)
            x_up = block(features)
            if i < len(self.up) - 1:
                x_up = F.interpolate(x_up, scale_factor=2.0, mode='nearest')
        return self.final_conv(x_up)
